searchData={"items":[{"type":"behaviour","title":"Pluggable","doc":"The step specification.\n\nThere are two kind of steps: function steps and module steps.\n\n##","ref":"Pluggable.html"},{"type":"behaviour","title":"Function steps - Pluggable","doc":"A function step is any function that receives a token and a set of\noptions and returns a token. Its type signature must be:\n\n    (Pluggable.Token.t, Pluggable.opts) :: Pluggable.Token.t\n\n##","ref":"Pluggable.html#module-function-steps"},{"type":"behaviour","title":"Module steps - Pluggable","doc":"A module step is an extension of the function step. It is a module that must\nexport:\n\n  * a `c:call/2` function with the signature defined above\n  * an `c:init/1` function which takes a set of options and initializes it.\n\nThe result returned by `c:init/1` is passed as second argument to `c:call/2`. Note\nthat `c:init/1` may be called during compilation and as such it must not return\npids, ports or values that are specific to the runtime.\n\nThe API expected by a module step is defined as a behaviour by the\n`Pluggable` module (this module).","ref":"Pluggable.html#module-module-steps"},{"type":"behaviour","title":"Examples - Pluggable","doc":"Here's an example of a function step:\n\n    def json_header_step(token, _opts) do\n      My.Token.put_data(token, \"some_data\")\n    end\n\nHere's an example of a module step:\n\n    defmodule PutSomeData do\n      def init(opts) do\n        opts\n      end\n\n      def call(token, _opts) do\n        My.Token.put_data(token, \"some_data\")\n      end\n    end","ref":"Pluggable.html#module-examples"},{"type":"behaviour","title":"The Pluggable Step pipeline - Pluggable","doc":"The `Pluggable.StepBuilder` module provides conveniences for building\npluggable step pipelines.","ref":"Pluggable.html#module-the-pluggable-step-pipeline"},{"type":"callback","title":"Pluggable.call/2","doc":"","ref":"Pluggable.html#c:call/2"},{"type":"callback","title":"Pluggable.init/1","doc":"","ref":"Pluggable.html#c:init/1"},{"type":"function","title":"Pluggable.run/3","doc":"Run a series of pluggable steps at runtime.\n\nThe steps given here can be either a tuple, representing a module step\nand their options, or a simple function that receives a token and\nreturns a token.\n\nIf any of the steps halt, the remaining steps are not invoked. If the\ngiven token was already halted, none of the steps are invoked\neither.\n\nWhile `Pluggable.StepBuilder` works at compile-time, this is a\nstraight-forward alternative that works at runtime.","ref":"Pluggable.html#run/3"},{"type":"function","title":"Examples - Pluggable.run/3","doc":"Pluggable.run(token, [{My.Step, []}, &IO.inspect/1])","ref":"Pluggable.html#run/3-examples"},{"type":"function","title":"Options - Pluggable.run/3","doc":"* `:log_on_halt` - a log level to be used if a pipeline halts","ref":"Pluggable.html#run/3-options"},{"type":"type","title":"Pluggable.opts/0","doc":"","ref":"Pluggable.html#t:opts/0"},{"type":"module","title":"Pluggable.PipelineBuilder","doc":"Build pluggable steps as pipelines of other steps. Use this only if you need\nto define and run multiple distinct pipelines in the same module.","ref":"Pluggable.PipelineBuilder.html"},{"type":"module","title":"Examples - Pluggable.PipelineBuilder","doc":"defmodule MyPipeline do\n      pipeline :foo do\n        plug SomePluggableStep\n        plug :inline_step\n      end\n\n      pipeline :bar do\n        plug AnotherPluggableStep\n        plug :inline_step\n      end\n    end\n\nThese pipelines can be run from within the same module:\n\n    Pluggable.run(token, [&foo(&1, [])])\n    Pluggable.run(another_token, [&bar(&1, [])])\n\nOr they can be run from outside\n\n    Pluggable.run(token, [&MyPipeline.foo(&1, [])])\n    Pluggable.run(another_token, [&MyPipeline.bar(&1, [])])","ref":"Pluggable.PipelineBuilder.html#module-examples"},{"type":"macro","title":"Pluggable.PipelineBuilder.pipeline/2","doc":"Defines a pluggable step as a pipeline of other steps.\n\nSee module doc for more information.","ref":"Pluggable.PipelineBuilder.html#pipeline/2"},{"type":"macro","title":"Pluggable.PipelineBuilder.step/2","doc":"Defines a step inside a pipeline.\n\nSee module doc for more information.","ref":"Pluggable.PipelineBuilder.html#step/2"},{"type":"module","title":"Pluggable.StepBuilder","doc":"Conveniences for building pipelines.\n\n    defmodule MyApp do\n      use Pluggable.StepBuilder\n\n      step SomeLibrary.Logger\n      step :hello, upper: true\n\n      # A function from another module can be plugged too, provided it's\n      # imported into the current module first.\n      import AnotherModule, only: [interesting_step: 2]\n      step :interesting_step\n\n      def hello(token, opts) do\n        body = if opts[:upper], do: \"WORLD\", else: \"world\"\n        send_resp(token, 200, body)\n      end\n    end\n\nMultiple steps can be defined with the `step/2` macro, forming a pipeline.\nThe steps in the pipeline will be executed in the order they've been added\nthrough the `step/2` macro. In the example above, `SomeLibrary.Logger` will\nbe called first and then the `:hello` function step will be called on the\nresulting token.","ref":"Pluggable.StepBuilder.html"},{"type":"module","title":"Options - Pluggable.StepBuilder","doc":"When used, the following options are accepted by `Pluggable.StepBuilder`:\n\n  * `:init_mode` - the environment to initialize the step's options, one of\n    `:compile` or `:runtime`. Defaults `:compile`.\n\n  * `:log_on_halt` - accepts the level to log whenever the request is halted\n\n  * `:copy_opts_to_assign` - an `atom` representing an assign. When supplied,\n    it will copy the options given to the Step initialization to the given\n    token assign","ref":"Pluggable.StepBuilder.html#module-options"},{"type":"module","title":"step behaviour - Pluggable.StepBuilder","doc":"Internally, `Pluggable.StepBuilder` implements the `Pluggable` behaviour, which\nmeans both the `init/1` and `call/2` functions are defined.\n\nBy implementing the Pluggable API, `Pluggable.StepBuilder` guarantees this module\nis a  pluggable step and can be run via `Pluggable.run/3` or used as part of\nanother pipeline.","ref":"Pluggable.StepBuilder.html#module-step-behaviour"},{"type":"module","title":"Overriding the default Pluggable API functions - Pluggable.StepBuilder","doc":"Both the `init/1` and `call/2` functions defined by `Pluggable.StepBuilder` can\nbe manually overridden. For example, the `init/1` function provided by\n`Pluggable.StepBuilder` returns the options that it receives as an argument, but\nits behaviour can be customized:\n\n    defmodule StepWithCustomOptions do\n      use Pluggable.StepBuilder\n      step SomeLibrary.Logger\n\n      def init(opts) do\n        opts\n      end\n    end\n\nThe `call/2` function that `Pluggable.StepBuilder` provides is used internally to\nexecute all the steps listed using the `step` macro, so overriding the\n`call/2` function generally implies using `super` in order to still call the\nstep chain:\n\n    defmodule StepWithCustomCall do\n      use Pluggable.StepBuilder\n      step SomeLibrary.Logger\n      step SomeLibrary.AddMeta\n\n      def call(token, opts) do\n        token\n        |> super(opts) # calls SomeLibrary.Logger and SomeLibrary.AddMeta\n        |> assign(:called_all_steps, true)\n      end\n    end","ref":"Pluggable.StepBuilder.html#module-overriding-the-default-pluggable-api-functions"},{"type":"module","title":"Halting a pluggable step pipeline - Pluggable.StepBuilder","doc":"A pluggable step pipeline can be halted with `Pluggable.Token.halt/1`. The builder\nwill prevent further steps downstream from being invoked and return the\ncurrent token. In the following example, the `SomeLibrary.Logger` step never\ngets called:\n\n    defmodule StepUsingHalt do\n      use Pluggable.StepBuilder\n\n      step :stopper\n      step SomeLibrary.Logger\n\n      def stopper(token, _opts) do\n        halt(token)\n      end\n    end","ref":"Pluggable.StepBuilder.html#module-halting-a-pluggable-step-pipeline"},{"type":"function","title":"Pluggable.StepBuilder.compile/3","doc":"Compiles a pluggable step pipeline.\n\nEach element of the pluggable step pipeline (according to the type signature of this\nfunction) has the form:\n\n    {step_name, options, guards}\n\nNote that this function expects a reversed pipeline (with the last step that\nhas to be called coming first in the pipeline).\n\nThe function returns a tuple with the first element being a quoted reference\nto the token and the second element being the compiled quoted pipeline.","ref":"Pluggable.StepBuilder.html#compile/3"},{"type":"function","title":"Examples - Pluggable.StepBuilder.compile/3","doc":"Pluggable.StepBuilder.compile(env, [\n      {SomeLibrary.Logger, [], true}, # no guards, as added by the Pluggable.StepBuilder.step/2 macro\n      {SomeLibrary.AddMeta, [], quote(do: a when is_binary(a))}\n    ], [])","ref":"Pluggable.StepBuilder.html#compile/3-examples"},{"type":"macro","title":"Pluggable.StepBuilder.step/2","doc":"A macro that stores a new step. `opts` will be passed unchanged to the new\nstep.\n\nThis macro doesn't add any guards when adding the new step to the pipeline;\nfor more information about adding steps with guards see `compile/3`.","ref":"Pluggable.StepBuilder.html#step/2"},{"type":"macro","title":"Examples - Pluggable.StepBuilder.step/2","doc":"step SomeLibrary.Logger               # step module\n    step :foo, some_options: true  # step function","ref":"Pluggable.StepBuilder.html#step/2-examples"},{"type":"type","title":"Pluggable.StepBuilder.step/0","doc":"","ref":"Pluggable.StepBuilder.html#t:step/0"},{"type":"protocol","title":"Pluggable.Token","doc":"Token protocol to be used with the `Pluggable` and `Pluggable.Builder` modules.\n\nWhen implementing pipelines using this library, a token holds the state and is\npassed on from step to step within the pipeline.","ref":"Pluggable.Token.html"},{"type":"protocol","title":"Deriving Pluggable.Token - Pluggable.Token","doc":"The simplest way to use this library is to define a token module which derives\n`Pluggable.Token` and defines a struct which, among others defines the keys:\n\n * `:halted` - the boolean status on whether the pipeline was halted\n * `:assigns` - shared user data as a map\n\n Example:\n\n    defmodule MyPipeline.Token do\n      @derive Pluggable.Token\n      defstruct [\n        halted: false,\n        assigns: %{},\n        # other state\n      ]\n    end\n\nIf the fields holding these two states are named differently, pass the fields\nas options to `@derive`:\n\n    defmodule MyPipeline.Token do\n      @derive {Pluggable.Token, halted_key: :stopped, assigns_key: :shared_state}\n      defstruct [\n        stopped: false,\n        shared_state: %{},\n        # other state\n      ]\n    end","ref":"Pluggable.Token.html#module-deriving-pluggable-token"},{"type":"protocol","title":"Implementing Pluggable.Token - Pluggable.Token","doc":"`Pluggable.Token` can be implemented. The following is the default implementation\nwhen deriving `Pluggable.Token`\n\n    defmodule MyPipeline.Token do\n      defstruct [\n        halted: nil,\n        assigns: %{},\n        # other state\n      ]\n    end\n\n    defimpl Pluggable.Token, for: MyPipeline.Token do\n      def halted?(token), do: token.halted\n\n      def halt(token), do: %{token | halted: true}\n\n      def assign(%MyPipeline.Token{assigns: assigns} = token, key, value) when is_atom(key) do\n        %{token | assigns: Map.put(assigns, key, value)}\n      end\n    end","ref":"Pluggable.Token.html#module-implementing-pluggable-token"},{"type":"function","title":"Pluggable.Token.assign/3","doc":"Assigns a value to a key in the shared user data map.","ref":"Pluggable.Token.html#assign/3"},{"type":"function","title":"Pluggable.Token.halt/1","doc":"Halts the Pluggable pipeline by preventing further steps downstream from being\ninvoked. See the docs for `Pluggable.Builder` for more information on halting a\nPluggable pipeline.","ref":"Pluggable.Token.html#halt/1"},{"type":"function","title":"Pluggable.Token.halted?/1","doc":"Returns the boolean status on whether the pipeline was halted","ref":"Pluggable.Token.html#halted?/1"},{"type":"type","title":"Pluggable.Token.t/0","doc":"All the types that implement this protocol.","ref":"Pluggable.Token.html#t:t/0"},{"type":"extras","title":"Pluggable","doc":"# Pluggable\n\n[![Module Version](https://img.shields.io/hexpm/v/pluggable.svg)](https://hex.pm/packages/pluggable)\n[![Coverage Status](https://coveralls.io/repos/github/mruoss/pluggable/badge.svg?branch=main)](https://coveralls.io/github/mruoss/pluggable?branch=main)\n[![Last Updated](https://img.shields.io/github/last-commit/mruoss/pluggable.svg)](https://github.com/mruoss/pluggable/commits/main)\n\n[![Build Status CI](https://github.com/mruoss/pluggable/actions/workflows/ci.yaml/badge.svg)](https://github.com/mruoss/pluggable/actions/workflows/ci.yaml)\n[![Build Status Elixir](https://github.com/mruoss/pluggable/actions/workflows/elixir_matrix.yaml/badge.svg)](https://github.com/mruoss/pluggable/actions/workflows/elixir_matrix.yaml)\n\n[![Hex Docs](https://img.shields.io/badge/hex-docs-lightgreen.svg)](https://hexdocs.pm/pluggable/)\n[![Total Download](https://img.shields.io/hexpm/dt/pluggable.svg)](https://hex.pm/packages/pluggable)\n[![License](https://img.shields.io/hexpm/l/pluggable.svg)](https://github.com/mruoss/pluggable/blob/main/LICENSE)\n\nPluggable helps to define `Plug` like pipelines but with arbitrary tokens.\nThe library comes with almost exact copies of the module `Plug` and\n`Plug.Builder`. However, instead of passing around a `%Plug.Conn{}` struct,\nthis library passes around a Token you define in your project.","ref":"readme.html"},{"type":"extras","title":"Credits - Pluggable","doc":"Most of the code in this module was copied from the\n[`:plug`](https://github.com/elixir-plug/plug/) library so credits go to the\ncreators and maintainers of `:plug`.","ref":"readme.html#credits"},{"type":"extras","title":"Installation - Pluggable","doc":"The package can be installed by adding `pluggable` to your list of dependencies\nin `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:pluggable, \"~> 1.0\"}\n  ]\nend\n```","ref":"readme.html#installation"},{"type":"extras","title":"Usage - Pluggable","doc":"To use this library, you first have to define your token. Once that\nis done, use `Pluggable.StepBuilder` to build steps and pipelines.\n\n#","ref":"readme.html#usage"},{"type":"extras","title":"Deriving Pluggable.Token - Pluggable","doc":"The easiest way to define a token is to create a module which derives\n`Pluggable.Token` and defines a struct which, among others defines the keys:\n\n- `:halted` - the boolean status on whether the pipeline was halted\n- `:assigns` - shared user data as a map\n\nExample:\n\n```elixir\ndefmodule MyPipeline.Token do\n  @derive Pluggable.Token\n  defstruct [\n    halted: false,\n    assigns: %{},\n    # other state\n  ]\nend\n```\n\nIf the fields holding these two states are named differently, pass the fields\nas options to `@derive`:\n\n```elixir\ndefmodule MyPipeline.Token do\n  @derive {Pluggable.Token, halted_key: :stopped, assigns_key: :shared_state}\n  defstruct [\n    stopped: false,\n    shared_state: %{},\n    # other state\n  ]\nend\n```\n\n#","ref":"readme.html#deriving-pluggable-token"},{"type":"extras","title":"Implementing Pluggable.Token - Pluggable","doc":"`Pluggable.Token` can be implemented. The following is the default implementation\nwhen deriving `Pluggable.Token`\n\n```elixir\n  defmodule MyPipeline.Token do\n    defstruct [\n      halted: nil,\n      assigns: %{},\n      # other state\n    ]\n  end\n\n  defimpl Pluggable.Token, for: MyPipeline.Token do\n    def halted?(token), do: token.halted\n\n    def halt(token), do: %{token | halted: true}\n\n    def assign(%MyPipeline.Token{assigns: assigns} = token, key, value) when is_atom(key) do\n      %{token | assigns: Map.put(assigns, key, value)}\n    end\n  end\n```","ref":"readme.html#implementing-pluggable-token"},{"type":"extras","title":"Building Pipelines - Pluggable","doc":"`Pluggable.StepBuilder` works just like `Plug.Builder`. See the\nmodule documentation for instructions.\n\n`Pluggable.PipelineBuilder` can be used to define and run multiple pipelines in \nthe same module. See the module documentation for instructions.","ref":"readme.html#building-pipelines"},{"type":"extras","title":"Code Formatting - Pluggable","doc":"When using the `Pluggable.StepBuilder`, you might want to format the usage\nof the `step` macro without parens. To configure the formatter not to add\nparens, add this to your `.formatter.exs`:\n\n```elixir\n# .formatter.exs\n[\n  import_deps: [:pluggable]\n]\n```","ref":"readme.html#code-formatting"},{"type":"extras","title":"Changelog","doc":"# Changelog\n\nAll notable changes to this project will be documented in this file.\n\nThe format is based on [Keep a Changelog](https://keepachangelog.com/en/1.0.0/),\nand this project adheres to [Semantic Versioning](https://semver.org/spec/v2.0.0.html).","ref":"changelog.html"},{"type":"extras","title":"Unreleased - Changelog","doc":"<!-- Add your changelog entry to the relevant subsection -->\n\n<!-- #","ref":"changelog.html#unreleased"},{"type":"extras","title":"Added | Changed | Deprecated | Removed | Fixed | Security --> - Changelog","doc":"<!-- No new entries below this line! -->\n\n## [1.1.0] - 2024-03-21","ref":"changelog.html#added-changed-deprecated-removed-fixed-security"},{"type":"extras","title":"Added - Changelog","doc":"- Add `Pluggable.PipelineBuilder` to create multiple pipelines in the same module ([#58](https://github.com/mruoss/pluggable/pull/58))\n\n## [1.0.1] - 2022-10-25\n\n#","ref":"changelog.html#added"},{"type":"extras","title":"Added - Changelog","doc":"- :locals_without_parens export in formatter configuration\n\n## [1.0.0] - 2022-10-12\n\nThis is the first release.","ref":"changelog.html#added"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,49,46,50]}}